/* send RTP packet Test program. V2.02.02 2007/10/11                        */
/* Copyright  2007 - 2008 Antonio Cattaneo, Valerio Messina                */
/* Original version: Antonio Cattaneo. Ported on Posix by Valerio Messina   
   Open socket in connection mode, to use 'send'.                           
   you can download last version on http://users.iol.it/efa                 */
/* sendRtp.c is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   sendRtp.c is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.          */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>                                                 /* errno */
#ifdef G_OS_WIN32                                        /* we are on Win32 */
#  include <windows.h>                                        /* Win32 only */
#  include <winsock.h>                                        /* Win32 only */
#  include <conio.h>                                          /* Win32 only */
#  include <winbase.h>                                        /* Win32 only */
#else
#  include <sys/socket.h> /* AF_INET, PF_INET, SOCK_DGRAM, SOL_SOCKET, SO_REUSEADDR */
                      /* socket, inet_addr, setsockopt, bind, connect, send */
#  include <netinet/in.h>         /* INADDR_ANY, htons, htonl, ntohs, ntohl */
#  include <arpa/inet.h>                                       /* inet_addr */
#  define __USE_POSIX199309    /* needed for timespec & nanosleep in time.h */
/*#  define __USE_XOPEN_EXTENDED */        /* needed for usleep in unistd.h */
#  include <unistd.h>     /* close as in SVr4, SVID, POSIX, X/OPEN, BSD 4.3 */
#endif
#include <time.h>
#include <sys/time.h>                       /* gettimeofday, struct timeval */
#include <limits.h>                                            /* LLONG_MAX */
#include <stdbool.h>                                       /* bool datatype */

#define BUFFER_SIZE   2048
#define MAX_CH        768 /* numero massimo di connessioni dell'aggregatore */
#define MAX_TIME      170       /* massima dimensione che sta nell'ethernet */
#define CAMP_PER_MSEC 8              /* numero di campioni per millisecondo */
#define PCMA          0x08                                  /* Payload Type */
#define HOST_ADDR     "10.2.0.140"                  /* indirizzo di default */
#define START_PORT    36866  /* 32770 primo vero canale audio delle piastre */
#ifdef G_OS_WIN32                                        /* we are on Win32 */
#  define INET_ADDRSTRLEN 16     /* su Win32 manca len ("xxx.xxx.xxx.xxx"); */
#endif

/* simple wrapper for Win32/POSIX 1003.1-2003 socket portability */
#ifndef G_OS_WIN32                           /* we are on POSIX 1003.1-2003 */
#  define SOCKET_ERROR_CODE() errno
#  define SOCKET_CLOSE(fd) close (fd)
#  define SOCKET_ERROR_IS_EINPROGRESS() (errno==EINPROGRESS)
#  define SOCKET_ERROR_IS_EINTR() (errno==EINTR)
#else                                                    /* we are on Win32 */
#  define SOCKET_ERROR_CODE() WSAGetLastError()
#  define SOCKET_CLOSE(fd) closesocket(fd)
#  define SOCKET_ERROR_IS_EINPROGRESS() (WSAGetLastError()==WSAEWOULDBLOCK)
#  define SOCKET_ERROR_IS_EINTR() 0                   /* No WSAEINTR errors */
#endif

bool rtpDbg = true;                         /* 'false' disable debug printf */

/* RTP packet defined with one contribute only */
struct rtpPacket {
   unsigned char  cc     :4;
   unsigned char  x      :1;
   unsigned char  p      :1;
   unsigned char  version:2;
   unsigned char  pt     :7;
   unsigned char  m      :1;
   unsigned short seqNum;
   unsigned int   timeStamp;
   unsigned int   ssrc;
   unsigned char  dataBuffer[BUFFER_SIZE];
} sendPkt[MAX_CH];

/* vettore di socket da aprire, una per canale */
#ifdef G_OS_WIN32                                        /* we are on Win32 */
   SOCKET sock[MAX_CH];                              /* 'SOCKET' Win32 only */
#else                                        /* we are on POSIX 1003.1-2003 */
   int sock[MAX_CH];
#endif

char ipAddr[INET_ADDRSTRLEN] = HOST_ADDR;
unsigned short int udpPrt = START_PORT;

/* socket, setsockopt, bind & connect of numCh channel using:
   sa=HOST_ADDR, UDP, port[START_PORT+numCh*2 : START_PORT], step 2.
   Return 0 OK, 1 on ERROR */
int openSock ( int numCh ) {
   int i;
   int err;
   int optVal = 1;
   struct sockaddr_in localAdd, remoteAdd;

   localAdd.sin_family       = AF_INET;
   localAdd.sin_addr.s_addr  = htonl ( INADDR_ANY );

   remoteAdd.sin_family      = AF_INET;
   remoteAdd.sin_addr.s_addr = inet_addr ( HOST_ADDR );

   for ( i = 0; i < numCh; i++ ) {
      sock[i] = socket ( PF_INET, SOCK_DGRAM, 0);
#ifdef G_OS_WIN32                                        /* we are on Win32 */
      if ( sock[i] == INVALID_SOCKET ) {     /* 'INVALID_SOCKET' win32 only */
#else
      if ( sock[i] == -1 ) {
#endif
/*       printf ( "Errore %d nell'apertura del socket\n", WSAGetLastError () ); *//* 'WSAGetLastError' Win32 only */
         printf ( "Errore %d nell'apertura del socket\n", SOCKET_ERROR_CODE () );
         return ( 1 );
      }

      /* Socket option per riutilizzo socket */
      err = setsockopt ( sock[i], SOL_SOCKET, SO_REUSEADDR, (char *)&optVal, sizeof (optVal) );
      if ( err == -1 ) {
          printf ( "Errore %d setsockopt\n", SOCKET_ERROR_CODE () );
          return ( 1 );
      }

      /* Apro tutte le porte pari a partire da START_PORT */
      localAdd.sin_port = htons ( udpPrt + i * 2 );
      err = bind ( sock[i], (struct sockaddr *)&localAdd, sizeof ( localAdd ) );
      if ( err == -1 ) {
          printf ( "Errore %d nella bind del local address\n", SOCKET_ERROR_CODE () );
          return ( 1 );
      }

      /* Connetto alle porte pari a partire da START_PORT */
      remoteAdd.sin_port = htons ( udpPrt + i * 2 );
/*    remoteAdd.sin_port = htons( 35840 ); */
      err = connect ( sock[i], (struct sockaddr *)&remoteAdd, sizeof ( remoteAdd ) );
      if ( err < 0 ) {
         printf ( "Errore %d nella connect\n", SOCKET_ERROR_CODE () );
         return ( 1 );
      }
   }
   return ( 0 );
}

/* close sockets of numCh channel. Return 0 OK, 1 on ERROR */
int closeSock ( int numCh ) {
   int i;
   int err;

   for ( i = 0; i < numCh; i++ ) {
/*    err = closesocket ( sock[i] ); */         /* 'closesocket' Win32 only */
      err = SOCKET_CLOSE ( sock[i] );
      if ( err < 0 ) {
         printf ( "Errore %d nella chiusura del socket\n", SOCKET_ERROR_CODE () );
         return ( 1 );
      }
   }
   return ( 0 );
}

/* per sendPkt[chNum] canali, prepara pacchetti RTP, PT=PCMA con:
   SeqNum, Timestamp, SSRC casuali,
   e payload sinusoidale a 2 KHz di lunghezza(durata)=pktTime*CAMP_PER_MSEC */
void rtpStaticSetup ( int chNum, int pktTime ) {
   int i;
   int byteCnt;

   for ( i = 0; i < chNum; i++ ) {
      sendPkt[i].version   = 2;
      sendPkt[i].p         = 0;
      sendPkt[i].x         = 0;
      sendPkt[i].cc        = 0;
      sendPkt[i].m         = 0;
      sendPkt[i].pt        = PCMA;
#ifdef BORLANDC                                        /* we use BorlandC++ */
      sendPkt[i].seqNum    = htons ( rand()   ); /* on Win32 return a short */
      sendPkt[i].timeStamp = htonl ( _lrand() );     /* '_lrand' win32 only */
      sendPkt[i].ssrc      = htonl ( _lrand() );     /* '_lrand' win32 only */
#else                               /* we are on Win32 or POSIX 1003.1-2003 */
      if (rtpDbg) printf ("RAND_MAX:%d\n", RAND_MAX);         /* 2147483647 */
      sendPkt[i].seqNum    = htons ( 0xFFFF * (long long)rand() / RAND_MAX );
      sendPkt[i].timeStamp = htonl ( rand() );
      sendPkt[i].ssrc      = htonl ( rand() );
#endif
      // Generazione della testata del payload RTP: 0xAA 0xBB 0xCC 0xDD
/*    sendPkt[i].dataBuffer[0] = 0xAA;
      sendPkt[i].dataBuffer[1] = 0xBB;
      sendPkt[i].dataBuffer[2] = 0xCC;
      sendPkt[i].dataBuffer[3] = 0xDD; */
      for ( byteCnt = 0; byteCnt < pktTime * CAMP_PER_MSEC; byteCnt++ ) {
/*        if (byteCnt < 42)
             sendPkt[i].dataBuffer[byteCnt] = 0;
          else
             sendPkt[i].dataBuffer[byteCnt] = 0xFF; */

          /* Simulo la generazione di una 2 KHz */
          if ( ( byteCnt % 4 ) < 2 )
             sendPkt[i].dataBuffer[byteCnt] = 0x8A;
          else
             sendPkt[i].dataBuffer[byteCnt] = 0x0A;
      }
   }
}

/* calcola e scrive i nuovi seqNum e timeStamp su chNum canali usando pktTime */
inline void rtpDynamicSetup ( int chNum, int pktTime ) {
   int i;

   for ( i = 0; i < chNum; i++ ) {
      sendPkt[i].seqNum    = htons ( ntohs ( sendPkt[i].seqNum ) + 1);
      sendPkt[i].timeStamp = htonl ( ntohl ( sendPkt[i].timeStamp ) + pktTime*CAMP_PER_MSEC);
   }
}

/* invia un pacchetto su tutti i chNum canali. Return 1 on ERROR, 0 on OK */
inline int sendAllCh ( int chNum, int pktTime ) {
   int   i;
   int   err;
   void* tmp;

   for ( i = 0; i < chNum; i++ ) {

      /* Salto i canali da 32 a 63 */
      /* if ( ( i % 64 ) >= 32 )
            continue;*/

/*    (struct rtpPacket *)tmp = &sendPkt[i]; *//* warning: use of cast expressions as lvalues is deprecated */
      tmp = &sendPkt[i];       /* indirizzo del pacchetto del canale iesimo */
/*    err = send ( sock[i], (const char *)tmp, pktTime * CAMP_PER_MSEC + 12, 0 );*/
      err = send ( sock[i], tmp, pktTime*CAMP_PER_MSEC + 12, 0 );
      if ( err == -1 ) {
         printf ( "Errore %d nella send\n", SOCKET_ERROR_CODE () );
         closeSock ( chNum );
         return ( 1 );
      }
   }
   return ( 0 );
}

/* Uso: sendRtp <numero canali> <numero loop=packet/canale> <lunghezza pacchetto [ms]> */
int main ( int argc, char *argv[] ) {
#ifdef G_OS_WIN32                                        /* we are on Win32 */
   int     err;
   WORD    wVersionRequested;                          /* 'WORD' Win32 only */
   WSADATA wsaData;                                 /* 'WSADATA' Win32 only */
   clock_t startClock, endClock, sleepClock;         /* signed su Win32 ??? */
   clock_t totalStartClock, totalStopClock;          /* signed su Win32 ??? */
#else                                        /* we are on POSIX 1003.1-2003 */
   struct timespec wait, rem;
   int ret;
   struct timeval startTime, endTime, sleepTime;
   struct timeval totalStartTime, totalStopTime;
   long long microStartTime, microEndTime, microSleepTime;
   long long microTotalStartTime, microTotalStopTime;
#endif
   int     chNum;
   int     loopNum;
   int     pktTime;
   int     loopCnt;
   char*   ptr;

#ifdef G_OS_WIN32                                        /* we are on Win32 */
   wVersionRequested = MAKEWORD ( 2, 0 ); //winsock ver2 /* 'MAKEWORD' Win32 only */
   err = WSAStartup ( wVersionRequested, &wsaData ); /* 'WSAStartup' Win32 only */
   if ( err != 0 ) {
      printf ("Nessuna winsock.dll trovata!\n");
      return ( 1 );
   }
#endif

   if (argc) argc--; /* ora argv[1] e' il primo argomento, escludendo 'sendRtp' */
   if (rtpDbg) printf ("modo DEBUG attivo\n");
   if (rtpDbg) printf ("argc:%u\n", argc);
   if ( (argc < 3) || (argc > 5) ) {
      printf ("\n");
      printf ( "sendRtp V2.02.02 2007/10/11 send RTP packets. By A.Cattaneo & V.Messina\n");
      printf ("\n");
      printf ( "Use: sendRtp chanNumber packetNumber packetLength[ms] <ipAddress <udpPort>>\n");
      printf ("\n");
      printf ("With no <ipAddress> will send to: %s\n", HOST_ADDR);
      printf ("With no <udpPort>   will send to: %u\n", START_PORT);
      printf ("\n");
      return ( -1 );
   }

   chNum   = atoi ( argv[1] ); if ( chNum > MAX_CH ) chNum = MAX_CH;
   loopNum = atoi ( argv[2] );
   pktTime = atoi ( argv[3] ); if ( pktTime > MAX_TIME ) pktTime = MAX_TIME;
   printf ( "Test on %d channels with %d packets, packet length %d [ms]\n",
            chNum, loopNum, pktTime );
   if (argc>3) {
      if (rtpDbg) printf ( "IP Destination Address: \"%s\"\n", argv[4]);
      ptr = strncpy (ipAddr, argv[4], INET_ADDRSTRLEN);
      if (ptr==NULL) strncpy (ipAddr, HOST_ADDR, INET_ADDRSTRLEN);
   }
   if (argc==5) {
      udpPrt = atoi (argv[5]);
      if (udpPrt==0) udpPrt = START_PORT;
   }
   printf ( "IP Destination Address: %s, UDP Destination Port: %u\n",
            ipAddr, udpPrt );

   printf ("Set real time priority and press Return ...");
   getchar ();
#ifdef G_OS_WIN32
   Sleep ( 1000 );                                    /* 'Sleep' Win32 only */
#else                                        /* we are on POSIX 1003.1-2003 */
   wait.tv_sec  = 1;
   wait.tv_nsec = 0;
   ret = nanosleep ( &wait, &rem );
   if (ret==-1) {
      printf ("nanosleep failed\n");
      return ( -1 );
   }
#endif

   srand ( time(NULL) );             /* Inizializzo seme per numeri casuali */

   if ( openSock ( chNum ) )
      return ( -1 );

   rtpStaticSetup ( chNum, pktTime );
   if (rtpDbg) printf ("LONG_MAX:%ld\n", LONG_MAX);        /* 2147483647 */
   if (rtpDbg) printf ("LLONG_MAX:%lld\n", LLONG_MAX); /* 9223372036854775807 */

   printf ("Start sending ...\n");
#ifdef G_OS_WIN32                                        /* we are on Win32 */
   totalStartClock = clock ();                /* tick assoluti all'ingresso */
#else                                        /* we are on POSIX 1003.1-2003 */
   ret = gettimeofday (&totalStartTime, 0);   /* tick assoluti all'ingresso */
   if (ret==-1) {
      printf ("gettimeofday failed\n");
      return ( -1 );
   }
   microTotalStartTime = totalStartTime.tv_sec*1E6 + totalStartTime.tv_usec;
#endif
   for ( loopCnt = 0; loopCnt < loopNum; loopCnt++ ) { /* numero di packet/canale */
      #ifdef G_OS_WIN32                                  /* we are on Win32 */
         startClock = clock();         /* tick assoluti all'inizio del loop */
      #else                                  /* we are on POSIX 1003.1-2003 */
         ret = gettimeofday (&startTime, 0); /* tick assoluti all'inizio del loop */
         if (ret==-1) {
            printf ("gettimeofday failed\n");
            return ( -1 );
         }
         microStartTime = startTime.tv_sec*1E6 + startTime.tv_usec;
      #endif

      if ( sendAllCh ( chNum, pktTime ) ) return ( -1 );

      rtpDynamicSetup ( chNum, pktTime );

      #ifdef G_OS_WIN32                                  /* we are on Win32 */
         endClock = clock();            /* tick assoluti alla fine del loop */
      #else                                  /* we are on POSIX 1003.1-2003 */
         ret = gettimeofday (&endTime, 0); /* tick assoluti alla fine del loop */
         if (ret==-1) {
            printf ("gettimeofday failed\n");
            return ( -1 );
         }
         microEndTime = endTime.tv_sec*1E6 + endTime.tv_usec;
      #endif

      #ifdef G_OS_WIN32                                  /* we are on Win32 */
         if ( endClock >= startClock )
            sleepClock = pktTime + startClock - endClock;
         else
            sleepClock = pktTime + startClock - (LONG_MAX - endClock);
         if ( sleepClock < 0 ) sleepClock = 0;
      #else                                  /* we are on POSIX 1003.1-2003 */
         if ( microEndTime>=microStartTime) { // OK
            microSleepTime = pktTime*1E3 + microStartTime - microEndTime;
            sleepTime.tv_sec  = microSleepTime/1E6; // OK
            sleepTime.tv_usec = microSleepTime - sleepTime.tv_sec*1E6;
         } else {
            microSleepTime = pktTime*1E3 + microStartTime - (LLONG_MAX - microEndTime);
            sleepTime.tv_sec  = microSleepTime/1E6; // OK
            sleepTime.tv_usec = microSleepTime - sleepTime.tv_sec*1E6; // OK
         }

         if ( microSleepTime<0 ) { // OK
            sleepTime.tv_sec  = 0;
            sleepTime.tv_usec = 0;
         }
      #endif

#ifdef G_OS_WIN32                                        /* we are on Win32 */
      Sleep ( sleepClock );                           /* 'Sleep' Win32 only */
#else                                        /* we are on POSIX 1003.1-2003 */
      /*usleep ( sleepTime*1000 );*//* This function is obsolete. Use nanosleep(2) */
      wait.tv_sec  = sleepTime.tv_sec;
      wait.tv_nsec = sleepTime.tv_usec*1E3;
      ret = nanosleep ( &wait, &rem );
      if (ret==-1) {
         printf ("nanosleep failed\n");
         return ( -1 );
      }
#endif

   }                                                     /* for ( loopCnt ) */

#ifdef G_OS_WIN32                                        /* we are on Win32 */
   totalStopClock = clock();                    /* tick assoluti all'uscita */
#else                                        /* we are on POSIX 1003.1-2003 */
   ret = gettimeofday (&totalStopTime, 0);      /* tick assoluti all'uscita */
   if (ret==-1) {
      printf ("gettimeofday failed\n");
      return ( -1 );
   }
   microTotalStopTime = totalStopTime.tv_sec*1E6 + totalStopTime.tv_usec;
#endif
   printf ("Stop sending.\n");
   if ( closeSock ( chNum ) )
      return ( -1 );

#ifdef G_OS_WIN32                                        /* we are on Win32 */
   if (rtpDbg) printf ("totalStartClock:%ld totalStopClock:%ld\n", totalStartClock, totalStopClock);
   if ( totalStopClock >= totalStartClock )
      printf ( "Transmitted %d RTPpackets of %d byte on %d channels with period %f [ms]\n",
               loopNum, pktTime * CAMP_PER_MSEC + 12, chNum,
               1.0 * (totalStopClock - totalStartClock) / loopNum );
   else
      printf ( "Transmitted %d RTPpackets of %d byte on %d channels with period %f [ms]\n",
               loopNum, pktTime * CAMP_PER_MSEC + 12, chNum,
               1.0 * (LONG_MAX - totalStopClock + totalStartClock) / loopNum );
#else                                        /* we are on POSIX 1003.1-2003 */
   if (rtpDbg) printf ("totalStartTime:%ld.%ld totalStopTime:%ld.%ld\n", totalStartTime.tv_sec, totalStartTime.tv_usec, totalStopTime.tv_sec, totalStopTime.tv_usec);
   if (rtpDbg) printf ("microTotalStartTime:%lld, microTotalStopTime:%lld\n", microTotalStartTime, microTotalStopTime);
   
   if ( microTotalStopTime >= microTotalStartTime ) // OK
      printf ( "Transmitted %d RTPpackets of %d byte on %d channels with period %f [ms]\n",
               loopNum, pktTime * CAMP_PER_MSEC + 12, chNum,
               1.0 * (microTotalStopTime - microTotalStartTime) / loopNum / 1E3); // OK
   else
      printf ( "Transmitted %d RTPpackets of %d byte on %d channels with period %f [ms]\n",
               loopNum, pktTime * CAMP_PER_MSEC + 12, chNum,
               1.0 * (LLONG_MAX - microTotalStopTime + microTotalStartTime) / loopNum / 1E3); // OK
#endif
   return ( 0 );
}

